#!/bin/sh

EXEC="./synap.exe"

# Set time limit for all operations
#ulimit -t 30

globallog=testall.log
rm -f $globallog
error=0
globalerror=0
numfail=0
numsuccess=0

keep=0

Usage() {
    echo "Usage: testall.sh [options] [.syn files]"
    echo "-k    Keep intermediate files"
    echo "-h    Print this help"
    exit 1
}

SignalError() {
    if [ $error -eq 0 ] ; then
	echo "FAILED"
	error=1
    fi
    echo "  $1"
}

# Compare <outfile> <reffile> <difffile>
# Compares the outfile with reffile.  Differences, if any, written to difffile
Compare() {
    generatedfiles="$generatedfiles $3"
    echo diff -b $1 $2 ">" $3 1>&2
    diff -b "$1" "$2" > "$3" 2>&1 || {
	SignalError "$1 differs"
	echo "FAILED $1 differs from $2" 1>&2
    }
}

# Run <args>
# Report the command, run it, and report any errors
Run() {
    echo "Begin Running \"$*\""
    #echo $* 1>&2
    eval $* || {
	SignalError "$1 failed on $*"
	return 1
    }
    echo "Finished Running \"$1\""
}

Check() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.syn//'`
    reffile=`echo $1 | sed 's/.syn$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    while read line
    do
	    ARGS=$line
	    break
    done < "../tests/${basename}.args"

    generatedfiles="${basename}.out" &&
    Run "$EXEC $ARGS" "<" $1 &&
    Compare ../tests/out.tmp ${reffile}.out ${basename}.out.diff

    # Report the status and clean up the generated files

    if [ $error -eq 0 ] ; then
	if [ $keep -eq 0 ] ; then
	    rm -f $generatedfiles
	fi
	echo "OK"
	echo "###### SUCCESS" 1>&2
	numsuccess=`expr $numsuccess + 1`
    else
	echo "###### FAILED" 1>&2
	globalerror=$error
	numfail=`expr $numfail + 1`
    fi
}
CheckFail() {
    error=0
    basename=`echo $1 | sed 's/.*\\///
                             s/.syn//'`
    reffile=`echo $1 | sed 's/.syn$//'`
    basedir="`echo $1 | sed 's/\/[^\/]*$//'`/."

    echo -n "$basename..."

    echo 1>&2
    echo "###### Testing $basename" 1>&2

    while read line
    do
	    ARGS=$line
	    break
    done < "../tests/${basename}.args"

    generatedfiles="${basename}.out" &&
    Run "$EXEC $ARGS" "<" $1

    # Report the status and clean up the generated files

    if [ $error -eq 0 ] ; then
	echo "###### FAILED" 1>&2
	numfail=`expr $numfail + 1`
    else
	echo "###### SUCCESS" 1>&2
	globalerror=$error
	numsuccess=`expr $numsuccess + 1`
    fi
}

while getopts kdpsh c; do
    case $c in
	k) # Keep intermediate files
	    keep=1
	    ;;
	h) # Help
	    Usage
	    ;;
    esac
done

shift `expr $OPTIND - 1`

if [ $# -ge 1 ]
then
    files=$@
else
    files="../tests/fail-*.syn ../tests/test-*.syn"
fi

for file in $files
do
    case $file in
	*test-*)
	    Check $file 2>> $globallog
	    ;;
	*fail-*)
	    CheckFail $file 2>> $globallog
	    ;;
	*)
	    echo "unknown file type $file"
	    globalerror=1
	    ;;
    esac
done

cat $globallog
total=`expr $numfail + $numsuccess`
echo "$numfail/$total tests failed"
exit $globalerror
